use Link;
use searchers::Searcher;

use hyper;
use hyper::Client;

use serde_json;
use serde_json::Value;

use retry;
use retry::retry;
use retry::delay::Fixed;

use std;
use std::io::Read;

//use std::any::Any;
use mopa::Any;

const NUM_RETRIES: usize = 10;
const RETRY_WAIT_MILLIS: u64 = 500;

#[derive(Debug)]
pub struct MtgCard {
    pub name: String,
    pub cost: String,
    pub typeline: String,
    pub rules: String,
    pub flavor: Option<String>,
    pub power: Option<String>,
    pub toughness: Option<String>,
    pub url: String,
    pub image_url: String
}

impl Link for MtgCard {
    fn label (&self) -> &str {
        &self.name
    }

    fn url (&self) -> &str {
        &self.url
    }

    fn as_any (&self) -> &Any {
        self
    }
}

pub struct MtgSearcher {
    client: Client
}

impl MtgSearcher {
    pub fn new () -> MtgSearcher {
        MtgSearcher {
            client: Client::new()
        }
    }

    fn do_search (&self, name: &str) -> Result<String, Error> {
        let mut contents = String::new();
        let api_url = &format!("https://api.magicthegathering.io/v1/cards?name={}", name);
        let mut response = retry(Fixed::from_millis(RETRY_WAIT_MILLIS).take(NUM_RETRIES), || {
            self.client.get(api_url).send()
        })?;
        response.read_to_string(&mut contents)?;
        Result::Ok(contents)
    }
}

fn parse_entry (page: String) -> Result<MtgCard, Error> {
    let parsed: Value = serde_json::from_str(&page)?;
    for entry in parsed["cards"].as_array().unwrap_or(&vec![]) {
        if entry["imageUrl"].as_str().is_some() {
            let card_name = entry["name"].as_str().map(String::from).expect("expected name in json data");
            let card_url = format!("http://magiccards.info/query?q=!{}", card_name);

            return Result::Ok(MtgCard {
                name: card_name,
                cost: entry["manaCost"].as_str().map(String::from).expect("expected cost in json data"),
                typeline: entry["type"].as_str().map(String::from).expect("expected type in json data"),
                rules: entry["text"].as_str().map(String::from).expect("expected text in json data"),
                flavor: entry["flavor"].as_str().map(String::from),
                power: entry["power"].as_str().map(String::from),
                toughness: entry["toughness"].as_str().map(String::from),
                url: card_url,
                image_url: entry["imageUrl"].as_str().map(String::from).expect("expected image url in json data")
            });
        }
    }

    Result::Err(Error::Other(String::from("No card info found")))
}

impl Searcher<MtgCard> for MtgSearcher {
    fn exact_search (&self, name: &str) -> Option<MtgCard> {
        let search = format!(r#""{}""#, name);
        self.do_search(&search).and_then(parse_entry).ok()
    }
}

#[derive(Debug)]
enum Error {
    Http(hyper::error::Error),
    Io(std::io::Error),
    Json(serde_json::Error),
    Other(String)
}

impl From<hyper::error::Error> for Error {
    fn from (error: hyper::error::Error) -> Error {
        Error::Http(error)
    }
}

impl From<std::io::Error> for Error {
    fn from (error: std::io::Error) -> Error {
        Error::Io(error)
    }
}

impl From<serde_json::Error> for Error {
    fn from (error: serde_json::Error) -> Error {
        Error::Json(error)
    }
}

impl From<retry::Error<hyper::Error>> for Error {
    fn from (err: retry::Error<hyper::Error>) -> Error {
        match err {
            retry::Error::Operation { error, total_delay, tries } => {
                Error::Http(error)
            }
        }
    }
}
